from django.contrib.auth import authenticate
from django.views import View
from django.http import JsonResponse

from logging import getLogger
from json import loads as json_load

from .auth import JWTifier
from .settings import DjangoJwtAuthMiddlewareSettings

logger = getLogger(__name__)
local_settings = DjangoJwtAuthMiddlewareSettings()


def set_cookie(response, key, value, path="/"):
    response.set_cookie(
        key=key,
        value=value,
        path=path,
        samesite=local_settings.JWT_COOKIE_SAMESITE,
        httponly=local_settings.JWT_COOKIE_HTTPONLY,
        secure=local_settings.JWT_COOKIE_SECURE,
    )


class Login(View):
    def post(self, request):
        if (
            request.content_type == "application/json"
            and not request.user.is_authenticated
        ):
            json = json_load(request.body)
            username = json["username"]
            password = json["password"]
        else:
            username = request.POST.get("username")
            password = request.POST.get("password")

        user = (
            authenticate(username=username, password=password) or request.user
        )

        if user.is_authenticated:
            jwtifier = JWTifier(user)
            access, refresh = jwtifier.create_refresh()

            response = JsonResponse(
                {"message": "Authentication successful."}, status=200
            )

            set_cookie(
                response, local_settings.JWT_ACCESS_HEADER, f"Bearer {access}"
            )
            set_cookie(
                response,
                f"Refresh{local_settings.JWT_ACCESS_HEADER}",
                f"{refresh}",
                path=local_settings.JWT_REFRESH_PATH,
            )

            return response

        else:
            return JsonResponse(
                {"message": "Valid user credentials not provided."}, status=400
            )


class Logout(View):
    def post(self, request):
        if request.user.is_authenticated:
            response = JsonResponse({"message": "Session ended."}, status=200)
            if local_settings.JWT_LOGOUT_ACCESS:
                set_cookie(response, local_settings.JWT_ACCESS_HEADER, "")
            set_cookie(
                response,
                f"Refresh{local_settings.JWT_ACCESS_HEADER}",
                "",
                path=local_settings.JWT_REFRESH_PATH,
            )
            return response
        else:
            return JsonResponse(
                {"message": "Must be authenticated to log out."}, status=401
            )


class Refresh(View):
    def post(self, request):
        try:
            # check for a valid refresh token
            token_str = request.META.get(
                f"HTTP_REFRESH{local_settings.JWT_ACCESS_HEADER}".upper()
            ) or request.COOKIES.get(
                f"Refresh{local_settings.JWT_ACCESS_HEADER}"
            )

            token = JWTifier.token_from_str(token_str)
            logger.debug(token)
            assert token is not None
            assert token["user"].is_active

            jwtifier = JWTifier(token["user"])
            assert jwtifier.validate_token(token_str, "refresh")
            access = jwtifier.create_access()

            response = JsonResponse(
                {"message": "Access token refreshed."}, status=200
            )
            set_cookie(
                response, local_settings.JWT_ACCESS_HEADER, f"Bearer {access}"
            )
            return response
        except (AssertionError, ValueError):
            return JsonResponse(
                {"message": "Invalid refresh token."}, status=403
            )



from django.conf import settings
from appconf import AppConf

from logging import getLogger

logger = getLogger(__name__)


class DjangoJwtAuthMiddlewareSettings(AppConf):
    JWT_ACCESS_DURATION = 5  # minutes
    JWT_ACCESS_HEADER = "Authorization"
    JWT_ALGORITHM = "HS256"
    JWT_COOKIE_HTTPONLY = True
    JWT_COOKIE_SAMESITE = "Strict"
    JWT_COOKIE_SECURE = True
    JWT_LOGOUT_ACCESS = True
    JWT_REFRESH_DURATION = 14  # days
    JWT_REFRESH_PATH = "/jwt/refresh/"

    def configure_jwt_access_duration(self, value):
        return (
            value
            if not hasattr(settings, "JWT_ACCESS_DURATION")
            else settings.JWT_ACCESS_DURATION
        )

    def configure_jwt_access_header(self, value):
        return (
            value
            if not hasattr(settings, "JWT_ACCESS_HEADER")
            else settings.JWT_ACCESS_HEADER
        )

    def configure_jwt_algorithm(self, value):
        return (
            value
            if not hasattr(settings, "JWT_ALGORITHM")
            else settings.JWT_ALGORITHM
        )

    def configure_jwt_cookie_httponly(self, value):
        return (
            value
            if not hasattr(settings, "JWT_COOKIE_HTTPONLY")
            else settings.JWT_COOKIE_HTTPONLY
        )

    def configure_jwt_cookie_samesite(self, value):
        return (
            value
            if not hasattr(settings, "JWT_COOKIE_SAMESITE")
            else settings.JWT_COOKIE_SAMESITE
        )

    def configure_jwt_cookie_secure(self, value):
        return (
            value
            if not hasattr(settings, "JWT_COOKIE_SECURE")
            else settings.JWT_COOKIE_SECURE
        )

    def configure_jwt_logout_access(self, value):
        return (
            value
            if not hasattr(settings, "JWT_LOGOUT_ACCESS")
            else settings.JWT_LOGOUT_ACCESS
        )

    def configure_jwt_refresh_duration(self, value):
        return (
            value
            if not hasattr(settings, "JWT_REFRESH_DURATION")
            else settings.JWT_REFRESH_DURATION
        )

    def configure_jwt_refresh_path(self, value):
        return (
            value
            if not hasattr(settings, "JWT_REFRESH_PATH")
            else settings.JWT_REFRESH_PATH
        )

    def configure(self):
        logger.debug(
            f"django_jwt_auth_middleware.settings: {self.configured_data}"
        )
        return self.configured_data



from django.contrib.auth.backends import BaseBackend
from django.contrib.auth.models import AnonymousUser
from django.contrib.auth import get_user_model, authenticate
from django.utils.deprecation import MiddlewareMixin
from jwt.exceptions import DecodeError, ExpiredSignatureError
from django.utils import timezone
from django.conf import settings as django_settings
from jwt import decode as jwt_decode, encode as jwt_encode

from datetime import timedelta
from logging import getLogger
from secrets import token_hex

from .settings import DjangoJwtAuthMiddlewareSettings

logger = getLogger(__name__)
local_settings = DjangoJwtAuthMiddlewareSettings()


class JWTifier:
    def __init__(self, user):
        self.user = user

    @classmethod
    def __encode(cls, user, duration, tok_type):
        now = timezone.now()
        jwt = jwt_encode(
            {
                "iat": now,
                "exp": now + duration,
                # "exp": now,
                "user": str(user.id),
                "tok": tok_type,
                "sec": token_hex(16),
            },
            django_settings.SECRET_KEY,
            algorithm=local_settings.JWT_ALGORITHM,
        )
        return jwt

    def create_access(
        self, duration=timedelta(minutes=local_settings.JWT_ACCESS_DURATION)
    ):
        return self.__encode(self.user, duration, "access")

    def create_refresh(
        self, duration=timedelta(days=local_settings.JWT_REFRESH_DURATION)
    ):
        return self.create_access(), self.__encode(
            self.user, duration, "refresh"
        )

    @classmethod
    def __decode(cls, jwt):
        try:
            token = jwt_decode(
                jwt,
                django_settings.SECRET_KEY,
                algorithms=local_settings.JWT_ALGORITHM,
            )
            user = get_user_model().objects.get(
                pk=token["user"], is_active=True
            )
            token["user"] = user
            return token
        except (
            DecodeError,
            ExpiredSignatureError,
            get_user_model().DoesNotExist,
        ):
            return None

    def validate_token(self, jwt, tok_type):
        token = self.__decode(jwt) or {"user": None, "tok": None}
        return token["user"] == self.user and token["tok"] == tok_type

    @classmethod
    def user_from_token(cls, jwt):
        token = cls.__decode(jwt) or {"user": None}
        return token["user"]

    @classmethod
    def token_from_str(cls, jwt):
        return cls.__decode(jwt)


class JWTAuthenticationBackend(BaseBackend):
    def authenticate(self, request, token=None):
        try:
            auth_type, auth_str = token.split(" ")
            assert auth_type.lower() == "bearer"

            return JWTifier.user_from_token(auth_str)
        except (
            AssertionError,
            get_user_model().DoesNotExist,
            ExpiredSignatureError,
            DecodeError,
            ValueError,
        ):
            return None

    def get_user(self, user_id):
        try:
            return get_user_model().objects.get(pk=user_id, is_active=True)
        except get_user_model().DoesNotExist:
            return None


class JWTAuthenticationMiddleware(MiddlewareMixin):
    def process_request(self, request):
        if not request.user.is_authenticated and (
            f"HTTP_{local_settings.JWT_ACCESS_HEADER.upper()}" in request.META
            or local_settings.JWT_ACCESS_HEADER in request.COOKIES
        ):
            auth_token_str = request.META.get(
                f"HTTP_{local_settings.JWT_ACCESS_HEADER.upper()}"
            ) or request.COOKIES.get(local_settings.JWT_ACCESS_HEADER)
            request.user = (
                authenticate(request, token=auth_token_str) or AnonymousUser()
            )



from django.urls import path, include

from . import views


subpatterns = [
    path("login/", views.Login.as_view()),
    path("logout/", views.Logout.as_view()),
    path("refresh/", views.Refresh.as_view()),
]

urlpatterns = [path("jwt/", include(subpatterns))]
